/* ****************************************** */
/* Emu75: HP-75 software emulator for PC      */
/*                                            */
/* desas75.c   desassembler module for Emu75  */
/*                                            */
/* Version 2.5, J-F Garnier 1997-2017         */
/* ****************************************** */

/* desas75.c */
/* desassembleur HP75 (et HP83,85,87) */



#include <stdio.h>
#include <string.h>


/* fonctions extern */
extern int ch_label(long adr, char *s); /* trans.c */
extern int hextoint(char c);    /* monit.c */

static int arp=0, drp=0;
static int ndef=0;



/* ARP - 0x00-0x3f */
static int desas_arp(unsigned char *codes, char *ligne)
{
  arp=codes[0];
  sprintf(ligne,"ARP  %o",arp);

  return(1);
}

/* DRP - 0x40-0x7f */
static int desas_drp(unsigned char *codes, char *ligne)
{
  drp=codes[0]-64;
  sprintf(ligne,"DRP  %o",drp);

  return(1);
}


/* instr. 1 octet - 0x98-0x9f*/
static int desas_1(unsigned char *codes,char *ligne)
{
  char *mnemo[]= { "BIN","BCD","SAD","DCE","ICE","CLE","RTN","PAD" } ;

  strcpy(ligne,mnemo[codes[0]-152]);
  return(1);
}


/* pseudo op gto */
static int desas_gto(unsigned char *codes,char *ligne)
{
  unsigned int v;
  char name[80];

  v=codes[1]+codes[2]*256+1;
  if (ch_label(v,name)) {
    sprintf(ligne,"GTO  =%s",name);
  }
  else {
    sprintf(ligne,"GTO  =%04X",v);
  }

  return(3);
}

/* JSB  - 0xc6/0xce */
static int desas_jsb(unsigned char *codes,char *ligne)
{
  unsigned int v;
  char name[80];

  v=codes[1]+codes[2]*256;
  if (ch_label(v,name)==0) sprintf(name,"%04X",v);
  if (codes[0]==0xc6)
    sprintf(ligne,"JSB  X%o,=%s",arp,name);
  else {
    sprintf(ligne,"JSB  =%s",name);
    if (v==0x4661) ndef=1; // MELJSB
    if (v==0x466D) ndef=1; // SYSJSB
    if (v==0x4694) ndef=2; // ROMJSB
  }

  return(3);
}


/* nops - 0xd6/0xde */
/* these NOP opcodes are described in the Emu85 from E. Kaser */
static int desas_nop(unsigned char *codes,char *ligne)
{
  if (codes[0]==0xd6) {
    sprintf(ligne,"NOP3");
    return(3);
  }
  else {
    sprintf(ligne,"NOP1");
    return(1);
  }
}


/* JMP - 0xf0-0xff */
static int desas_jmp(unsigned int addr,unsigned  char *codes,char *ligne)
{
  char *mnemo[] = { "JMP","JNO","JOD","JEV","JNG","JPS","JNZ","JZR",
		    "JEN","JEZ","JNC","JCY","JLZ","JLN","JRZ","JRN" } ;
  int v;

  v=codes[1];
  if (v>127) v=v-256;
  v=v+addr+2;
  sprintf(ligne,"%s  %04X",mnemo[codes[0]-240],v);
  return(2);
}


/* cas general - 0x80-0xef */
static int desas_gen(unsigned char *codes,char *ligne)
{
  static char *mnemo[] = {
		    "EL d",  "ER d",  "LL d",  "LR d",
		    "IC d",  "DC d",  "TC d",  "NC d",
		    "TS d",  "CL d",  "OR da", "XR da",
		    "??",    "??",    "??",    "??",
		    "LD da", "ST da", "LDDda", "STDda",
		    "LD d=o","ST d=o","LDIda", "STIda",
		    "LDDd=v","STDd=v","LDDdxv","STDdxv",
		    "LDId=v","STId=v","LDIdxv","STIdxv",
		    "CM da", "AD da", "SB da", "AN da",
		    "CM d=o","AD d=o","SB d=o","AN d=o",
		    "CMDd=v","ADDd=v","SBDd=v","ANDd=v",
		    "CMDda", "ADDda", "SBDda", "ANDda",
		    "PODd+", "PODd-", "PUDd+", "PUDd-",
		    "POId+", "POId-", "PUId+", "PUId-" } ;
  char *ptr;
  char s[32];
  char name[80];
  int i, lg;
  int n;
  unsigned int v;

  ptr=mnemo[(codes[0]-128)/2];
  strncpy(ligne,ptr,2);
  ligne[0]=*ptr++;
  ligne[1]=*ptr++;
  if ((codes[0]&1)==0)
    ligne[2]='B';
  else
    ligne[2]='M';
  ligne[3]=0;
  while (*ptr!=0) {
    switch (*ptr) {
      case 'a':
//	sprintf(s,",R%o",arp);
	if (arp!=1) sprintf(s,",R%o",arp); else sprintf(s,",R*"); // 0.97b
	strcat(ligne,s);
	n=1;
	break;
      case '+':
	sprintf(s,",+R%o",arp);
	strcat(ligne,s);
	n=1;
	break;
      case '-':
	sprintf(s,",-R%o",arp);
	strcat(ligne,s);
	n=1;
	break;
      case 'x':
	sprintf(s,",X%o,",arp);
	strcat(ligne,s);
	n=1;
	break;
      case 'd':
	if (drp!=1) sprintf(s," R%o",drp); else sprintf(s," R*");
	strcat(ligne,s);
	n=1;
	break;
      case 'o':
	if ((codes[0]&1)==0)
	  lg=1;
	else
	  if (drp<32) lg=2-(drp%2); else lg=8-(drp%8);
        if (lg==2){
          v=codes[1]+codes[2]*256;
          if (ch_label(v,name))
            sprintf(s,"%s",name);
          else
            sprintf(s,"%04X",v);
          strcat(ligne,s);

        }
        else
          for (i=1;i<=lg;i++) {
            if (i>1) strcat(ligne,",");
            sprintf(s,"%02X",codes[i]);
            strcat(ligne,s);
          }
	n=1+lg;
	break;
      case 'v':
	v=codes[1]+codes[2]*256;
        if (ch_label(v,name))
          sprintf(s,"%s",name);
        else
          sprintf(s,"%04X",v);
	strcat(ligne,s);
	n=3;
	break;
      case '=':
	strcat(ligne,",=");
	break;
      default:
	s[0]=*ptr;
	s[1]=0;
	strcat(ligne,s);
    } /* end switch */
    ptr++;
  } /* end while */

  return(n);
}


static int desas_def(unsigned char *codes,char *ligne)
{
  unsigned int v;
  char name[80];

  v=codes[0]+codes[1]*256;
  if (ch_label(v,name)==0) sprintf(name,"%04X",v);
  sprintf(ligne,"DEF  =%s",name);
  return(2);
}



int desas(unsigned int addr,unsigned  char *codes, char *ligne)
{
  unsigned char c;
  int n;

  if (ndef) {
    n=desas_def(codes,ligne);
    ndef--;
  }
  else {
    c=codes[0];
    if (c<0x40) n=desas_arp(codes,ligne);
    else if (c<0x80) n=desas_drp(codes,ligne);
    else if ((c>=0x98)&&(c<=0x9f)) n=desas_1(codes,ligne);
    else if ((c==0xa9)&&(drp==4)) n=desas_gto(codes,ligne);
    else if ((c==0xc6)||(c==0xce)) n=desas_jsb(codes,ligne);
    else if ((c==0xd6)||(c==0xde)) n=desas_nop(codes,ligne);
    else if (c>=0xf0) n=desas_jmp(addr,codes,ligne);
    else n=desas_gen(codes,ligne);
  }

  return(n);
}


void setad(int a, int d)
{
  arp=a;
  drp=d;
}


/* historical note:
this disassembler is a modified version of a HP85 disassembler I wrote in the '80s,
then adapted to the HP75.
10 ! --- desas ---
20 ! desassembleur HP75
30 !
35 ON ERROR GOTO 1200
40 Y=PEEK(0)
45 OFF ERROR
50 H$="0123456789ABCDEF"
60 INPUT "adr debut (hex)";A$
70 I=FND(A$,16)
80 INPUT "adr fin (hex)";A$
90 I1=FND(A$,16)
100 !
200 A0$,D0$="0"
210 !
220 PRINT FNH$(I);TAB(6);
230 X=PEEK(I)
240 IF X<64 THEN GOSUB 360 @ GOTO 300
250 IF X<128 THEN GOSUB 410 @ GOTO 300
260 IF X>151 AND X<160 THEN GOSUB 800 @ GOTO 300
270 IF X=198 OR X=206 THEN GOSUB 720 @ GOTO 300
280 IF X>239 THEN GOSUB 610 @ GOTO 300
290 GOSUB 440
300 I=I+1
310 IF I<=I1 THEN 220
320 !
330 PRINT "--- fin ---"
340 END
350 ! ****************
360 ! ---arp
370 A0$=FNO$(X)
380 PRINT "ARP ";A0$
390 RETURN
400 ! ---drp
410 D0$=FNO$(X-64)
420 PRINT "DRP ";D0$
430 RETURN
440 ! cas general
450 RESTORE
460 FOR J=0 TO (X-128)/2
470 READ A$
480 NEXT J
490 PRINT A$[1,2];
500 IF MOD(X,2)=0 THEN PRINT "B";ELSE PRINT "M";
510 FOR J=3 TO LEN(A$)
520 B$=A$[J,J]
530 IF B$="a" THEN PRINT "R";A0$;" "; @ GOTO 580
540 IF B$="d" THEN PRINT " R";D0$;" "; @ GOTO 580
550 IF B$="o" THEN GOSUB 870 @ GOTO 580
560 IF B$="v" THEN GOSUB 970 @ GOTO 580
570 PRINT B$;
580 NEXT J
590 PRINT
600 RETURN
610 ! Jump
620 RESTORE 1160
630 FOR J=0 TO X-240
640 READ A$
650 NEXT J
660 PRINT A$;" ";
670 I=I+1
680 Y=PEEK(I)
690 IF Y>127 THEN Y=Y-256
700 PRINT FNH$(I+Y+1)
710 RETURN
720 ! jsb
730 PRINT "JSB ";
740 IF X=198 THEN PRINT "XR";A0$;
750 IF X=206 THEN PRINT "=";
760 Y=PEEK(I+1)+PEEK(I+2)*256
770 PRINT " ";FNH$(Y)
780 I=I+2
790 RETURN
800 ! instr 1 octet
810 RESTORE 1180
820 FOR J=0 TO X-152
830 READ A$
840 NEXT J
850 PRINT A$
860 RETURN
870 ! cas =octets
880 IF MOD(X,2)=0 THEN L=1 @ GOTO 910
890 Y=FND(D0$,8)      !!!
900 IF Y<32 THEN L=2-MOD(Y,2) ELSE L=8-MOD(Y,8)
910 FOR J0=1 TO L
920 I=I+1
930 Y=PEEK(I)
940 PRINT " ";FNH$(Y);
950 NEXT J0
960 RETURN
970 ! cas =vecteur
980 Y=PEEK(I+1)+PEEK(I+2)*256
990 I=I+2
1000 PRINT " ";FNH$(Y);
1010 RETURN
1020 DATA ELd,ERd,LLd,LRd
1030 DATA ICd,DCd,TCd,NCd
1040 DATA TSd,CLd,ORda,XRda
1050 DATA x,x,x,x
1060 DATA LDda,STda,LDDda,STDda
1070 DATA LDd=o,STd=o,LDIda,STIda
1080 DATA LDDd=v,STDd=v,LDDdXav,STDdXav
1090 DATA LDId=v,STId=v,LDIdXav,STIdXav
1100 DATA CMda,ADda,SBda,ANda
1110 DATA CMd=o,ADd=o,SBd=o,ANd=o
1120 DATA CMDd=v,ADDd=v,SBDd=v,ANDd=v
1130 DATA CMDda,ADDda,SBDda,ANDda
1140 DATA PODd+a,PODd-a,PUDd+a,PUDd-a
1150 DATA POId+a,POId-a,PUId+a,PUId-a
1160 DATA JMP,JNO,JOD,JEV,JNG,JPS,JNZ,JZR
1170 DATA JEN,JEZ,JNC,JCY,JLZ,JLN,JRZ,JRN
1180 DATA BIN,BCD,SAD,DCE,ICE,CLE,RTN,PAD
1190 !
1200 OFF ERROR
1210 DISP "ERROR: memxlex needed"
1220 STOP
1230 !
1240 DEF FND(A$,N)
1250 X=0
1260 FOR K=1 TO LEN(A$)
1270 X=X*N+POS(H$,A$[K,K])-1
1280 NEXT K
1290 FND=X
1300 END DEF
1310 !
1320 DEF FNH$(X)
1330 A$=""
1340 N=MOD(X,16)+1
1350 A$=H$[N,N]&A$
1360 X=X\16 @ IF X<>0 THEN 1340
1370 FNH$=A$
1380 END DEF
1390 !
1400 DEF FNO$(X)
1410 A$=""
1420 N=MOD(X,8)+1
1430 A$=H$[N,N]&A$
1440 X=X\8 @ IF X<>0 THEN 1420
1450 FNO$=A$
1460 END DEF
*/
